@@ -1,6 +1,9 @@
 import gym
 from gym import spaces
 import copy
+
+from numpy.core._multiarray_umath import ndarray
+
 from gym_craftingworld.envs.rendering import *
 import matplotlib.pyplot as plt
 import matplotlib.animation as animation
@@ -39,6 +42,7 @@ TASK_LIST = ['MakeBread', 'EatBread', 'BuildHouse', 'ChopTree', 'ChopRock', 'GoT
 class CraftingWorldEnvRGB(gym.GoalEnv):
     """Custom Crafting environment that follows the gym interface pattern
     """
+
     metadata = {'render.modes': ['human', 'Non']}
 
     def __init__(self, size=(10, 10), fixed_init_state=None, fixed_goal=None, tasks_to_ignore=None, store_gif=False,
@@ -58,6 +62,7 @@ class CraftingWorldEnvRGB(gym.GoalEnv):
         self.metadata = {'render.modes': ['human', 'Non']}
 
         self.num_rows, self.num_cols = size
+        self.agent_start = (int(self.num_rows/2), int(self.num_cols/2))
 
         self.max_steps = max_steps
 
@@ -75,7 +80,11 @@ class CraftingWorldEnvRGB(gym.GoalEnv):
                                                                           dtype=int),
                                                   achieved_goal=spaces.Box(low=0, high=255, shape=(self.num_rows*4,
                                                                                                    self.num_cols*4, 3),
-                                                                           dtype=int)))
+                                                                           dtype=int),
+                                                  init_observation=spaces.Box(low=0, high=255, shape=(self.num_rows * 4,
+                                                                                                 self.num_cols * 4, 3),
+                                                                         dtype=int)
+                                                  ))
 
         self.observation_vector_space = spaces.Dict(dict(observation=spaces.Box(low=0, high=1,
                                                                                 shape=(self.num_rows, self.num_cols,
@@ -87,7 +96,13 @@ class CraftingWorldEnvRGB(gym.GoalEnv):
                                                                                  dtype=int),
                                                          achieved_goal=spaces.Box(low=0, high=1,
                                                                                   shape=(1, len(self.task_list)),
-                                                                                  dtype=int)))
+                                                                                  dtype=int),
+                                                         init_observation=spaces.Box(low=0, high=1,
+                                                                                shape=(self.num_rows, self.num_cols,
+                                                                                       len(OBJECTS) + 1 + len(
+                                                                                           PICKUPABLE)),
+                                                                                dtype=int)
+                                                         ))
 
         # TODO: wrapper that flattens to regular env, wrapper that changes desired goal to dict of rewards, reward wrapper
 
@@ -116,14 +131,18 @@ class CraftingWorldEnvRGB(gym.GoalEnv):
 
         # self.observation = {'observation': self.obs_one_hot, 'desired_goal': self.desired_goal_vector,
         #                     'achieved_goal': self.achieved_goal_vector}
+        self.INIT_OBS_VECTOR = copy.deepcopy(self.obs_one_hot)
+        self.INIT_OBS = self.render(self.INIT_OBS_VECTOR)
         self.observation_vector = {'observation': self.obs_one_hot, 'desired_goal': self.desired_goal_vector,
-                                   'achieved_goal': self.achieved_goal_vector}
+                                   'achieved_goal': self.achieved_goal_vector,
+                                   'init_observation': self.INIT_OBS_VECTOR}
         # self.init_obs_one_hot = copy.deepcopy(self.obs_one_hot)
         self.init_observation_vector = copy.deepcopy(self.observation_vector)
 
         self.desired_goal = self.imagine_obs()
         self.observation = {'observation': self.render(self.obs_one_hot), 'desired_goal': self.desired_goal,
-                            'achieved_goal': self.render(self.obs_one_hot)}
+                            'achieved_goal': self.render(self.obs_one_hot),
+                            'init_observation': self.INIT_OBS}
         self.init_observation = copy.deepcopy(self.observation)
 
         self.ACTIONS = [coord(-1, 0, name='up'), coord(0, 1, name='right'), coord(1, 0, name='down'),
@@ -174,14 +193,18 @@ class CraftingWorldEnvRGB(gym.GoalEnv):
         else:
             self.obs_one_hot, self.agent_pos = self.sample_state()
 
+        self.INIT_OBS_VECTOR = copy.deepcopy(self.obs_one_hot)
+        self.INIT_OBS = self.render(self.INIT_OBS_VECTOR)
         self.observation_vector = {'observation': self.obs_one_hot, 'desired_goal': self.desired_goal_vector,
-                                   'achieved_goal': self.achieved_goal_vector}
+                                   'achieved_goal': self.achieved_goal_vector,
+                                   'init_observation': self.INIT_OBS_VECTOR}
 
         self.init_observation_vector = copy.deepcopy(self.observation_vector)
 
         self.desired_goal = self.imagine_obs()
         self.observation = {'observation': self.render(self.obs_one_hot), 'desired_goal': self.desired_goal,
-                            'achieved_goal': self.render(self.obs_one_hot)}
+                            'achieved_goal': self.render(self.obs_one_hot),
+                            'init_observation': self.INIT_OBS}
         self.init_observation = copy.deepcopy(self.observation)
 
         self.reward = self.calculate_rewards()
@@ -328,9 +351,9 @@ class CraftingWorldEnvRGB(gym.GoalEnv):
         task_success = self.eval_tasks()
         self.achieved_goal_vector = self.task_one_hot(task_success)
         self.observation_vector = {'observation': self.obs_one_hot, 'desired_goal': self.desired_goal_vector,
-                                   'achieved_goal': self.achieved_goal_vector}
+                                   'achieved_goal': self.achieved_goal_vector, 'init_observation': self.INIT_OBS_VECTOR}
         self.observation = {'observation': self.render(self.obs_one_hot), 'desired_goal': self.desired_goal,
-                            'achieved_goal': self.render(self.obs_one_hot)}
+                            'achieved_goal': self.render(self.obs_one_hot), 'init_observation': self.INIT_OBS}
         observation = self.observation
         self.reward = self.calculate_rewards()
         reward = self.reward
@@ -491,7 +514,7 @@ Desired Goals: {}""".format(self.ep_no, self.step_num, action_label, desired_goa
 
         self.ims.append([im, ttl, txt])
 
-    def sample_state(self):
+    def sample_state_not_fixed_agent(self):
         """
         produces a observation with one of each object
         :return obs: a sample observation
@@ -511,6 +534,29 @@ Desired Goals: {}""".format(self.ep_no, self.step_num, action_label, desired_goa
 
         return state, agent_position
 
+    def sample_state(self):
+        """
+        produces a observation with one of each object
+        :return obs: a sample observation
+        :return agent_position: position of the agent within the observation
+        """
+        objects = [self.one_hot(i - 1) for i in range(1,9)]
+        grid = objects + [[0 for _ in range(self.observation_vector_space.spaces['observation'].shape[2])]
+                          for _ in range(self.num_rows * self.num_cols - len(objects))]
+        random.shuffle(grid)
+
+        state = np.asarray(grid, dtype=int).reshape(self.observation_vector_space.spaces['observation'].shape)
+        while np.argmax(state[self.agent_start[0]][self.agent_start[1]]) in [3, 4, 5, 6]:
+            # don't start agent on rock, tree, house or bread
+            np.random.shuffle(state)
+        agent_encoding = self.one_hot(8)
+        state[self.agent_start[0]][self.agent_start[1]] += agent_encoding
+        agent_position = coord(self.agent_start[0],
+                               self.agent_start[1],
+                               self.num_rows - 1, self.num_cols - 1)
+
+        return state, agent_position
+
     def eval_tasks(self):
         # TODO: eval_tasks is not efficient (dictcomps are slow) - try to speed this up
         task_success = {}
@@ -538,10 +584,17 @@ Desired Goals: {}""".format(self.ep_no, self.step_num, action_label, desired_goa
             # print(task_success[task])
         return goal_one_hot
 
-    def calculate_rewards(self):
-        error = np.sum(np.square(self.observation['desired_goal'] - self.observation['achieved_goal']))
+    def calculate_rewards(self, desired_goal=None, achieved_goal=None, initial_goal=None):
+        if desired_goal is None:
+            desired_goal = self.observation['desired_goal']
+        if achieved_goal is None:
+            achieved_goal = self.observation['achieved_goal']
+        if initial_goal is None:
+            initial_goal = self.init_observation['achieved_goal']
+        error = np.sqrt(np.sum(np.square(desired_goal - achieved_goal)))
         if self.pos_rewards is True:
-            return np.sum(self.desired_goal_vector) - error
+            initial_error = np.sqrt(np.sum(np.square(desired_goal - initial_goal)))
+            return -error/initial_error
         return -error
 
     def get_objects(self, code, state):
@@ -595,3 +648,4 @@ Desired Goals: {}""".format(self.ep_no, self.step_num, action_label, desired_goa
         holding = np.argmax(one_hot_row[len(OBJECTS) + 1:]) if one_hot_row[len(OBJECTS) + 1:].any() == 1 else None
         agent = one_hot_row[len(OBJECTS)]
         return object_at_location, agent, holding
+
